
import ox.cads.experiments.{Experiments,Graphs}

/** An experiment comparing the three testers, on a map and a queue. */
object Experiment{
  // lengths of iterations per measurement to consider
  val MinLogNumIters = 5; val MaxLogNumIters = 14 // ***
  val numsIters = (MinLogNumIters to MaxLogNumIters).map(1 << _).toArray

  val TotalIters = 1 << 18 // total number of iterations per worker

  val ErrorSignal = ox.cads.experiments.Experiments.ErrorSignal

  // Array giving names of testers (for graphs) the appropriate flags, and the
  // log of the max number of iterations it can handle
  val mapTesters =  // ***
    Array(
      // Graph search
      ("WG-Graph", "--WGGraph"), 
      ("JIT-Graph", "--JITGraph"), 
      // ("Linear", "--linear"),
      // Competition; names give the tree search algorithm, then the graph 
      // search algorithm
      ("WG-WG-Competition", "--WG-WG-Comp"), 
      ("JIT-WG-Competition", "--JIT-WG-Comp"),
      ("WG-JIT-Competition",  "--WG-JIT-Comp"),
      ("JIT-JIT-Competition", "--JIT-JIT-Comp")
    )
  val queueTesters = Array(
    ("WG-Graph", "--WGGraph"), 
    ("JIT-Graph", "--JITGraph"), 
    ("Queue", "--QLT"),
    ("WG-WG-Competition", "--WG-WG-Comp"), 
    ("JIT-WG-Competition", "--JIT-WG-Comp"),
    ("WG-JIT-Competition",  "--WG-JIT-Comp"),
    ("JIT-JIT-Competition", "--JIT-JIT-Comp"),
    ("WG-Queue-Competition", "--WG-QLT-Comp"), 
    ("JIT-Queue-Competition", "--JIT-QLT-Comp")
  )


  // ("Linear", ""), ("Wing \\& Gong", "--WG"), ("JITLinUndo", "--JITLinUndo")

  val Billion = 1000000000

  /** Parameters defining statistical analysis of experiments: repeat until 95% 
    * confidence half-interval is at most 1% of mean or until 50 repetitions
    * done. */
  val params = new Experiments.Params(50, 0.05, 0.01) 
  // val params = new Experiments.Params(2, 0.05, 0.1) // ***
  // val params = new Experiments.Params(10, 0.05, 0.05) // ***

  def main(args: Array[String]) = {
    // List of datatypes to test: true = queue; false = map
    var dtTypes = List(false,true) 
    // Parse command line
    var i = 0
    while(i < args.length){
      if(args(i) == "--queueOnly"){dtTypes = List(true); i += 1 }
      else if(args(i) == "--mapOnly"){dtTypes = List(false); i += 1 }
      else{ println("Illegal usage"); sys.exit }
    }

    for(queue <- dtTypes){ 
      val testers = if(queue) queueTesters else mapTesters
      val prog = if(queue) "QueueTest" else "MapTest"
      val cmd0 = "scala -J-Xmx10g "+prog+" --timing "
      // Array for results, in millis
      val results = 
	Array.ofDim[(Double,Double)](testers.length, numsIters.length)
      for(i <- 0 until numsIters.length){
	val numIters = numsIters(i); val reps = TotalIters / numIters
	println("\nnumIters = "+numIters+"; reps = "+reps)
	// Each measurement will consider reps runs, in each of which each
	// worker performs numIters operations.
	for(t <- 0 until testers.length){
	  println(testers(t)._1+" "+prog)
	  // Avoid long runs if the previous length failed.
	  if(i > 0 && results(t)(i-1)._1 == ErrorSignal){
	    println("Skipping"); results(t)(i) = (ErrorSignal,0)
	  }
	  else{
	    val cmd = cmd0+testers(t)._2+" --iters "+numIters+" --reps "+reps
	    // A single measurement, giving time in nanos
	    // println(cmd)
	    def measure : Double = {
	      val time = Experiments.timeProc(cmd)
	      print(time/1000000+" "); time.toDouble
	    }
	    val (m, s) = Experiments.iterateMeasurement(measure, params)
            println(m/1000000,s/1000000) // in millis
	    if(m != ErrorSignal) results(t)(i) = (m/Billion,s/Billion)
	    else results(t)(i) = (ErrorSignal, 0)
	  }
	}
      } // end of outer for loop

      // ox.cads.util.Profiler.report; sys.exit // ****
      // Produce graphs
      val options = 
	Array[String](
          "title = Comparison of linearizability testers on a "+
	  (if(queue) "queue" else "map"), 
	  "xlabel = Number of operations per run", 
          "ylabel = Time (s)", 
          "log basis x=2", 
          "legend pos = north west",
          "height = 0.9\\textheight", "width = 0.9\\textwidth"
	)
      val graphString = 
	Graphs.makeLogXGraph(options, testers.map(_._1), numsIters, results)
      // println(graphString)
      val fname = if(queue) "queueTesting.tex" else "mapTesting.tex"
      Graphs.writeStandAloneFile(fname, graphString)
    }
  }
}
