import ox.cads.util.Profiler
import ox.cads.testing._
// import ox.cads.collection.ShardedMap
import scala.util.Random
import scala.reflect.ClassTag


/** A linearizabilty tester for a Map[A,A].
  * @param makeDatum a function that produces random data
  * @param default default datum 
  * @param p the number of workers
  * @param iters the number of operations performed by a worker in each run
  * @param reps the number of runs to perform; a negative value indicates to
  * run until an error is found
  * @param maxKey the size of the key space
  * @param concMapType a string describing the type of concurrent map to use
  * @param testerType a string indicating the linearizability algorithm to use.
  * @param quiet should the testing be silent?
  */

class MapLinTest[A: ClassTag](
  makeDatum: (Int, Random) => A, default: A, p: Int, iters: Int, 
  reps: Int, maxKey: Int,  concMapType: String, testerType: String, 
  tsLog: Boolean, quiet: Boolean)
{
  // Probabilities of different operations
  val (readProb, writeProb, gOEUProb, delProb) = (0.25, 0.25, 0.25, 0.25)
    // (0.5, 0.5, 0.0, 0.0)  (0.3, 0.3, 0.0, 0.4) 
  val MaxDatum = 4 
  // Maximum number of iterations to be performed by the WG Graph Search, the
  // JIT Graph search and a tree search algorithm, respectively.
  val Million = 1000000
  val MaxSizeWGG = 1300000 // 1.3 * 10^6
  val MaxSizeJITG = -1 // 20 * Million // 10000000 // 10 * 10^6
  val MaxSizeT = 10000L * Million // 10 billion 
  // FIXME: allow these to be set as arguments.

  assert(readProb + writeProb + gOEUProb + delProb == 1.0)

  // Types of concurrent and sequential maps
  type CMap = ox.cads.collection.HashMap[A, A]
  type SMap = scala.collection.immutable.Map[A, A]
  type USMap = UndoableMap[A, A] 

  // Operations upon the sequential maps
  def seqGetOrElse(key: A, datum: A)(m: SMap) : (A, SMap) = 
    (m.getOrElse(key, datum), m)
  def seqUpdate(key: A, datum: A)(m: SMap) : (Unit, SMap) =
    ((), m.updated(key, datum))
  def seqGetOrElseUpdate(key: A, datum: A)(m: SMap) : (A, SMap) =
    if(m.contains(key)) (m(key), m) else (datum, m.updated(key, datum))
  def seqDelete(key: A)(m: SMap) : (Unit, SMap) =  ((), m - key)

  /** A worker for testers based on an immutable sequential datatype. */
  def worker(me: Int, log: GenericThreadLog[SMap,CMap]) = {
    val random = new scala.util.Random
    for(i <- 0 until iters){
      val key = makeDatum(maxKey, random); val rand = random.nextFloat
      if(rand < readProb) // getOrElse
	log.log(_.getOrElse(key, default), 
		   "getOrElse("+key+", "+default+")", 
		   seqGetOrElse(key, default))
      else if(rand < readProb+writeProb){ // update
	val datum = makeDatum(MaxDatum, random)
	log.log(_.update(key, datum), 
		   "update("+key+", "+datum+")", seqUpdate(key, datum))
      }
      else if(rand < readProb+writeProb+gOEUProb){ // getOrElseUpdate
	val datum = makeDatum(MaxDatum, random)
	log.log(_.getOrElseUpdate(key, datum), 
		   "getOrElseUpdate("+key+", "+datum+")",
		   seqGetOrElseUpdate(key, datum))
      }
      else // delete
	log.log(_.delete(key), "delete("+key+")", seqDelete(key))
    }
  }

  /** A worker for the testers based on an undoable sequential datatype. */
  def uWorker(me: Int, log: GenericThreadLog[USMap, CMap]) = {
    val random = new scala.util.Random
    for(i <- 0 until iters){
      val key = makeDatum(maxKey, random); val rand = random.nextFloat
      if(rand < readProb) // getOrElse
        log.log(_.getOrElse(key, default), 
		   "getOrElse("+key+", "+default+")", 
		   _.getOrElse(key, default))
      else if(rand < readProb+writeProb){ // update
        val datum = makeDatum(MaxDatum, random)
        log.log(_.update(key, datum), 
                   "update("+key+", "+datum+")", _.update(key, datum))
      }
      else if(rand < readProb+writeProb+gOEUProb){ // getOrElseUpdate 
        val datum = makeDatum(MaxDatum, random)
        log.log(_.getOrElseUpdate(key, datum), 
                   "getOrElseUpdate("+key+", "+datum+")",
                   _.getOrElseUpdate(key, datum))
      }
      else // delete
        log.log(_.delete(key), "delete "+key, _.delete(key))
    }
  }

  type CompLog = CompetitionThreadLog[SMap, USMap, CMap] 
  /** A worker for a competition linearizability tester using an immutable
    * and an undoable sequential specification. */
  def compWorker(me: Int, log: CompLog) = {
    val random = new scala.util.Random
    for(i <- 0 until iters){
      val key = makeDatum(maxKey, random); val rand = random.nextFloat
      if(rand < readProb) // getOrElse
        log.log(_.getOrElse(key, default), "getOrElse("+key+", "+default+")", 
		seqGetOrElse(key, default), _.getOrElse(key, default))
      else if(rand < readProb+writeProb){ // update
        val datum = makeDatum(MaxDatum, random)
        log.log(_.update(key, datum), "update("+key+", "+datum+")", 
		seqUpdate(key, datum), _.update(key, datum))
      }
      else if(rand < readProb+writeProb+gOEUProb){ // getOrElseUpdate 
        val datum = makeDatum(MaxDatum, random)
        log.log(_.getOrElseUpdate(key, datum), 
		"getOrElseUpdate("+key+", "+datum+")",
                seqGetOrElseUpdate(key, datum), _.getOrElseUpdate(key, datum))
      }
      else // delete
        log.log(_.delete(key), "delete "+key, seqDelete(key), _.delete(key))
    }
  }

  /** Choose size to make LvdPW hash table */
  private def mkSize : Int = {
    // Estimate of the size needed
    val cap = ((iters * (writeProb + gOEUProb) * p) min maxKey) * 1.6 
    var s = 2; while(s < cap) s *= 2; s
  }

  /** Create sequential maps */
  def imSeqMap = scala.collection.immutable.Map[A, A]()
  def uSeqMap = new UndoableMap[A, A]

  /** Run the tester */
  def apply() : Int = {
    def concMap : CMap = concMapType match{
      case "Sharded" => new ox.cads.collection.ShardedMap[A, A]
      case "LvdPWFaulty" => new LvdPWFaultyMap[A,A](mkSize)
      case "LvdPWResizeMapFaulty" => new LvdPWResizeMapFaulty[A,A](4)
      case "LvdPWBadHash" => new LvdPWBadHashMap[A,A](4)
      case "LockFreeReadShardedFaulty" => 
	new LockFreeReadShardedMapFaulty[A,A](2,4)
      case "RSO" => new ox.cads.collection.RecursiveSplitOrderingMap[A,A](2)
      // case "LvdPWResize" => new LvdPWResizeMap[A, A](8)
      // case "SimpleSharded" => new SimpleShardedMap[A, A]
      // case "Sharded" => new ShardedMap[A, A]
      // case "LockFreeReadSharded" => new LockFreeReadShardedMap[A, A]
      // case "LockFreeReadAtomicSharded" => 
      //   new LockFreeReadAtomicShardedMap[A, A]
    }

    // The testers (sequential only)
    def WGTree = LinearizabilityTester.WGTree[USMap,CMap](
      uSeqMap, concMap, p, uWorker, iters, tsLog, MaxSizeT)
    def JITTree = LinearizabilityTester.JITTree[USMap, CMap](
      uSeqMap, concMap, p, uWorker, iters, tsLog, MaxSizeT)
    def WGGraph = LinearizabilityTester.WGGraph[SMap, CMap](
      imSeqMap, concMap, p, worker, iters, tsLog, MaxSizeWGG)
    def BFSJIT = LinearizabilityTester.BFSJIT[SMap, CMap](
      imSeqMap, concMap, p, worker, iters, tsLog, MaxSizeT)
    def JITGraph = LinearizabilityTester.JITGraph[SMap, CMap](
      imSeqMap, concMap, p, worker, iters, tsLog, MaxSizeJITG)
    // Competition testers; tree-search algorithm name comes first
    def WGWGComp = CompetitionTester.WGWG(
      compWorker, p, iters, concMap, imSeqMap, uSeqMap, 
      tsLog, MaxSizeT, MaxSizeWGG)
    def WGJITComp = CompetitionTester.WGJIT( // WG Tree, JIT Graph
      compWorker, p, iters, concMap, imSeqMap, uSeqMap, 
      tsLog, MaxSizeT, MaxSizeJITG)
    def JITWGComp = CompetitionTester.JITWG( // JIT Tree, WG Graph
      compWorker, p, iters, concMap, imSeqMap, uSeqMap, 
      tsLog, MaxSizeT, MaxSizeWGG)
    def JITJITComp = CompetitionTester.JITJIT(
      compWorker, p, iters, concMap, imSeqMap, uSeqMap, 
      tsLog, MaxSizeT, MaxSizeJITG)

    // Now run the test
    var r = 0L; var done = false
    var result = 2
    while(r != reps && result > 0){
      result = testerType match{
	// Tree searches
 	case "WGTree" => { val tester = WGTree; tester() }
	case "JITTree" => { val tester = JITTree; tester() }
	// Graph searches
	case "WGGraph" => { val tester = WGGraph; tester() }
	case "JITGraph" => { val tester = JITGraph; tester() }
	case "BFSJIT" => { val tester = BFSJIT; tester() }
	// Competition searches
	case "WG-WG-Comp" => { val tester = WGWGComp; tester() }
	case "WG-JIT-Comp" => { val tester = WGJITComp; tester() }
	case "JIT-WG-Comp" => { val tester = JITWGComp; tester() }
	case "JIT-JIT-Comp" => { val tester = JITJITComp; tester() }
      } // end of testerType match
      // if(done) println("!")

      if(!quiet && r%50 == 0) print(".")
      r += 1
    } // end of while loop
    // if(done) println("error found")
    result
  }
}
