/** A map with an undo operation, for use in the Wing & Gong Linearizability 
  * Tester. */

class UndoableMap[A,B] extends scala.collection.mutable.Undoable{
  private val map = scala.collection.mutable.Map[A,B]()

  // We store a stack of the following events to record what has to be done to
  // undo updates
  abstract private class UndoEvent
  private case class UpdateEvent(key: A, value: B) extends UndoEvent
  private case class DeleteEvent(key: A) extends UndoEvent
  private case object NullEvent extends UndoEvent

  // The stack storing undo information
  private val stack = new scala.collection.mutable.Stack[UndoEvent]()

  def getOrElse(key: A, default: => B) : B = {
    stack.push(NullEvent); map.getOrElse(key, default)
  }

  /** If k is not in the map, initialise it to the result of op; otherwise 
    * return the value it maps to. */
  def getOrElseUpdate(key: A, op: => B) : B = {
    map.get(key) match{
      case None => stack.push(DeleteEvent(key))
      case Some(v) => stack.push(UpdateEvent(key, v))
    }
    map.getOrElseUpdate(key, op)
  }

  /** Add key -> data to the mapping, overwriting any previous value there. */
  def update(key: A, datum: B) : Unit = {
    map.get(key) match{
      case None => stack.push(DeleteEvent(key))
      case Some(v) => stack.push(UpdateEvent(key, v))
    }
    map.update(key, datum)
  }

  /** Delete value associated with key */
  def delete(key: A) : Unit = {
    map.get(key) match{
      case None => stack.push(NullEvent)
      case Some(v) => stack.push(UpdateEvent(key, v))
    }
    map.remove(key)
  }

  def undo : Unit = {
    // if(stack.length > 100) println(stack.length)
    stack.pop match{
      case UpdateEvent(key, value) => map.update(key, value)
      case DeleteEvent(key) => map.remove(key)
      case NullEvent => { }
    }
  }
}

