/** A queue with an undo option */

class UndoableQueue[T] extends scala.collection.mutable.Undoable{
  private val buff = new scala.collection.mutable.ArrayBuffer[T]
  private var start = 0
  private var next = 0
  // This represents the queue with contents buff[start..next).  Undoing
  // dequeue will give access to the values in buff[0..start).
  // DTI: 0 <= start <= next <= buff.size

  // We keep objects of the following types in a stack to allow undoing. 
  private abstract class UndoEvent{ 
    def apply : Unit
  }
  private case class UndoEnqueue(x: T) extends UndoEvent{
    def apply = {
      assert(next>start && buff(next-1) == x); next -= 1
    }
  }
  private case class UndoDequeue(x: T) extends UndoEvent{
    def apply = {
      assert(start>0 && buff(start-1) == x); start -= 1
    }
  }
  private case object NullUndo extends UndoEvent{
    def apply = {}
  }

  private val undoStack = new scala.collection.mutable.Stack[UndoEvent]

  def enqueue(x: T) : Unit = {
    if(next == buff.size) buff += x else buff(next) = x
    next += 1
    undoStack.push(UndoEnqueue(x))
  }

  def dequeue : Option[T] = { 
    if(start==next){ undoStack.push(NullUndo); None }
    else{ 
      assert(start<next); val x = buff(start); start += 1;
      undoStack.push(UndoDequeue(x)); Some(x)
    }
  }

  def undo = undoStack.pop.apply

  override def toString = 
    buff.drop(start).take(next-start).mkString("<", ",", ">")
}
  
